@@ -1,7 +1,7 @@
 /*
  * libeio implementation
  *
- * Copyright (c) 2007,2008,2009,2010,2011 Marc Alexander Lehmann <libeio@schmorp.de>
+ * Copyright (c) 2007,2008 Marc Alexander Lehmann <libeio@schmorp.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without modifica-
@@ -38,10 +38,6 @@
  */
 
 #include "eio.h"
-
-#ifdef EIO_STACKSIZE
-# define XTHREAD_STACKSIZE EIO_STACKSIZE
-#endif
 #include "xthread.h"
 
 #include <errno.h>
@@ -51,7 +47,6 @@
 #include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/statvfs.h>
 #include <limits.h>
 #include <fcntl.h>
 #include <assert.h>
@@ -71,6 +66,7 @@
 #ifdef _WIN32
 
   /*doh*/
+
 #else
 
 # include "config.h"
@@ -81,28 +77,6 @@
 # include <signal.h>
 # include <dirent.h>
 
-#if _POSIX_MEMLOCK || _POSIX_MEMLOCK_RANGE || _POSIX_MAPPED_FILES
-# include <sys/mman.h>
-#endif
-
-/* POSIX_SOURCE is useless on bsd's, and XOPEN_SOURCE is unreliable there, too */
-# if __FreeBSD__ || defined __NetBSD__ || defined __OpenBSD__
-#  define _DIRENT_HAVE_D_TYPE /* sigh */
-#  define D_INO(de) (de)->d_fileno
-#  define D_NAMLEN(de) (de)->d_namlen
-# elif __linux || defined d_ino || _XOPEN_SOURCE >= 600
-#  define D_INO(de) (de)->d_ino
-# endif
-
-#ifdef _D_EXACT_NAMLEN
-# undef D_NAMLEN
-# define D_NAMLEN(de) _D_EXACT_NAMLEN (de)
-#endif
-
-# ifdef _DIRENT_HAVE_D_TYPE
-#  define D_TYPE(de) (de)->d_type
-# endif
-
 # ifndef EIO_STRUCT_DIRENT
 #  define EIO_STRUCT_DIRENT struct dirent
 # endif
@@ -112,38 +86,26 @@
 #if HAVE_SENDFILE
 # if __linux
 #  include <sys/sendfile.h>
-# elif __FreeBSD__ || defined __APPLE__
+# elif __freebsd
 #  include <sys/socket.h>
 #  include <sys/uio.h>
 # elif __hpux
 #  include <sys/socket.h>
-# elif __solaris
+# elif __solaris /* not yet */
 #  include <sys/sendfile.h>
 # else
 #  error sendfile support requested but not available
 # endif
 #endif
 
-#ifndef D_TYPE
-# define D_TYPE(de) 0
-#endif
-#ifndef D_INO
-# define D_INO(de) 0
-#endif
-#ifndef D_NAMLEN
-# define D_NAMLEN(de) strlen ((de)->d_name)
-#endif
+/* number of seconds after which an idle threads exit */
+#define IDLE_TIMEOUT 10
 
 /* used for struct dirent, AIX doesn't provide it */
 #ifndef NAME_MAX
 # define NAME_MAX 4096
 #endif
 
-/* used for readlink etc. */
-#ifndef PATH_MAX
-# define PATH_MAX 4096
-#endif
-
 /* buffer size for various temporary buffers */
 #define EIO_BUFSIZE 65536
 
@@ -195,7 +157,6 @@ static void eio_execute (struct etp_worker *self, eio_req *req);
       closedir (wrk->dirp);	\
       wrk->dirp = 0;		\
     }
-
 #define ETP_WORKER_COMMON \
   void *dbuf;	\
   DIR *dirp;
@@ -204,7 +165,7 @@ static void eio_execute (struct etp_worker *self, eio_req *req);
 
 #define ETP_NUM_PRI (ETP_PRI_MAX - ETP_PRI_MIN + 1)
 
-/* calculate time difference in ~1/EIO_TICKS of a second */
+/* calculcate time difference in ~1/EIO_TICKS of a second */
 static int tvdiff (struct timeval *tv1, struct timeval *tv2)
 {
   return  (tv2->tv_sec  - tv1->tv_sec ) * EIO_TICKS
@@ -222,13 +183,12 @@ static unsigned int max_poll_reqs;     /* reslock */
 static volatile unsigned int nreqs;    /* reqlock */
 static volatile unsigned int nready;   /* reqlock */
 static volatile unsigned int npending; /* reqlock */
-static volatile unsigned int max_idle = 4;      /* maximum number of threads that can idle indefinitely */
-static volatile unsigned int idle_timeout = 10; /* number of seconds after which an idle threads exit */
+static volatile unsigned int max_idle = 4;
 
-static xmutex_t wrklock;
-static xmutex_t reslock;
-static xmutex_t reqlock;
-static xcond_t  reqwait;
+static mutex_t wrklock = X_MUTEX_INIT;
+static mutex_t reslock = X_MUTEX_INIT;
+static mutex_t reqlock = X_MUTEX_INIT;
+static cond_t  reqwait = X_COND_INIT;
 
 #if !HAVE_PREADWRITE
 /*
@@ -236,7 +196,7 @@ static xcond_t  reqwait;
  * normal read/write by using a mutex. slows down execution a lot,
  * but that's your problem, not mine.
  */
-static xmutex_t preadwritelock = X_MUTEX_INIT;
+static mutex_t preadwritelock = X_MUTEX_INIT;
 #endif
 
 typedef struct etp_worker
@@ -244,7 +204,7 @@ typedef struct etp_worker
   /* locked by wrklock */
   struct etp_worker *prev, *next;
 
-  xthread_t tid;
+  thread_t tid;
 
   /* locked by reslock, reqlock or wrklock */
   ETP_REQ *req; /* currently processed request */
@@ -368,14 +328,6 @@ static ETP_REQ *reqq_shift (etp_reqq *q)
   abort ();
 }
 
-static void etp_thread_init (void)
-{
-  X_MUTEX_CREATE (wrklock);
-  X_MUTEX_CREATE (reslock);
-  X_MUTEX_CREATE (reqlock);
-  X_COND_CREATE  (reqwait);
-}
-
 static void etp_atfork_prepare (void)
 {
   X_LOCK (wrklock);
@@ -423,13 +375,12 @@ static void etp_atfork_child (void)
   nready   = 0;
   npending = 0;
 
-  etp_thread_init ();
+  etp_atfork_parent ();
 }
 
 static void
 etp_once_init (void)
-{
-  etp_thread_init ();
+{    
   X_THREAD_ATFORK (etp_atfork_prepare, etp_atfork_parent, etp_atfork_child);
 }
 
@@ -616,7 +567,7 @@ static void etp_submit (ETP_REQ *req)
 static void etp_set_max_poll_time (double nseconds)
 {
   if (WORDACCESS_UNSAFE) X_LOCK   (reslock);
-  max_poll_time = nseconds * EIO_TICKS;
+  max_poll_time = nseconds;
   if (WORDACCESS_UNSAFE) X_UNLOCK (reslock);
 }
 
@@ -630,14 +581,7 @@ static void etp_set_max_poll_reqs (unsigned int maxreqs)
 static void etp_set_max_idle (unsigned int nthreads)
 {
   if (WORDACCESS_UNSAFE) X_LOCK   (reqlock);
-  max_idle = nthreads;
-  if (WORDACCESS_UNSAFE) X_UNLOCK (reqlock);
-}
-
-static void etp_set_idle_timeout (unsigned int seconds)
-{
-  if (WORDACCESS_UNSAFE) X_LOCK   (reqlock);
-  idle_timeout = seconds;
+  max_idle = nthreads <= 0 ? 1 : nthreads;
   if (WORDACCESS_UNSAFE) X_UNLOCK (reqlock);
 }
 
@@ -775,11 +719,6 @@ void eio_set_max_idle (unsigned int nthreads)
   etp_set_max_idle (nthreads);
 }
 
-void eio_set_idle_timeout (unsigned int seconds)
-{
-  etp_set_idle_timeout (seconds);
-}
-
 void eio_set_min_parallel (unsigned int nthreads)
 {
   etp_set_min_parallel (nthreads);
@@ -799,8 +738,6 @@ int eio_poll (void)
 /* work around various missing functions */
 
 #if !HAVE_PREADWRITE
-# undef pread
-# undef pwrite
 # define pread  eio__pread
 # define pwrite eio__pwrite
 
@@ -830,17 +767,17 @@ eio__pwrite (int fd, void *buf, size_t count, off_t offset)
   ooffset = lseek (fd, 0, SEEK_CUR);
   lseek (fd, offset, SEEK_SET);
   res = write (fd, buf, count);
-  lseek (fd, ooffset, SEEK_SET);
+  lseek (fd, offset, SEEK_SET);
   X_UNLOCK (preadwritelock);
 
   return res;
 }
 #endif
 
-#ifndef HAVE_UTIMES
+#ifndef HAVE_FUTIMES
 
-# undef utimes
-# define utimes(path,times)  eio__utimes (path, times)
+# define utimes(path,times)  eio__utimes  (path, times)
+# define futimes(fd,times)   eio__futimes (fd, times)
 
 static int
 eio__utimes (const char *filename, const struct timeval times[2])
@@ -858,13 +795,6 @@ eio__utimes (const char *filename, const struct timeval times[2])
     return utime (filename, 0);
 }
 
-#endif
-
-#ifndef HAVE_FUTIMES
-
-# undef futimes
-# define futimes(fd,times) eio__futimes (fd, times)
-
 static int eio__futimes (int fd, const struct timeval tv[2])
 {
   errno = ENOSYS;
@@ -874,40 +804,10 @@ static int eio__futimes (int fd, const struct timeval tv[2])
 #endif
 
 #if !HAVE_FDATASYNC
-# undef fdatasync
-# define fdatasync(fd) fsync (fd)
+# define fdatasync fsync
 #endif
 
-/* sync_file_range always needs emulation */
-int
-eio__sync_file_range (int fd, off_t offset, size_t nbytes, unsigned int flags)
-{
-#if HAVE_SYNC_FILE_RANGE
-  int res;
-
-  if (EIO_SYNC_FILE_RANGE_WAIT_BEFORE   != SYNC_FILE_RANGE_WAIT_BEFORE
-      || EIO_SYNC_FILE_RANGE_WRITE      != SYNC_FILE_RANGE_WRITE
-      || EIO_SYNC_FILE_RANGE_WAIT_AFTER != SYNC_FILE_RANGE_WAIT_AFTER)
-    {
-      flags = 0
-         | (flags & EIO_SYNC_FILE_RANGE_WAIT_BEFORE ? SYNC_FILE_RANGE_WAIT_BEFORE : 0)
-         | (flags & EIO_SYNC_FILE_RANGE_WRITE       ? SYNC_FILE_RANGE_WRITE       : 0)
-         | (flags & EIO_SYNC_FILE_RANGE_WAIT_AFTER  ? SYNC_FILE_RANGE_WAIT_AFTER  : 0);
-    }
-
-  res = sync_file_range (fd, offset, nbytes, flags);
-
-  if (!res || errno != ENOSYS)
-    return res;
-#endif
-
-  /* even though we could play tricks with the flags, it's better to always
-   * call fdatasync, as that matches the expectation of its users best */
-  return fdatasync (fd);
-}
-
 #if !HAVE_READAHEAD
-# undef readahead
 # define readahead(fd,offset,count) eio__readahead (fd, offset, count, self)
 
 static ssize_t
@@ -944,7 +844,7 @@ eio__sendfile (int ofd, int ifd, off_t offset, size_t count, etp_worker *self)
 # if __linux
   res = sendfile (ofd, ifd, &offset, count);
 
-# elif __FreeBSD__
+# elif __freebsd
   /*
    * Of course, the freebsd sendfile is a dire hack with no thoughts
    * wasted on making it similar to other I/O functions.
@@ -953,27 +853,8 @@ eio__sendfile (int ofd, int ifd, off_t offset, size_t count, etp_worker *self)
     off_t sbytes;
     res = sendfile (ifd, ofd, offset, count, 0, &sbytes, 0);
 
-    #if 0 /* according to the manpage, this is correct, but broken behaviour */
-    /* freebsd' sendfile will return 0 on success */
-    /* freebsd 8 documents it as only setting *sbytes on EINTR and EAGAIN, but */
-    /* not on e.g. EIO or EPIPE - sounds broken */
-    if ((res < 0 && (errno == EAGAIN || errno == EINTR) && sbytes) || res == 0)
-      res = sbytes;
-    #endif
-
-    /* according to source inspection, this is correct, and useful behaviour */
-    if (sbytes)
-      res = sbytes;
-  }
-
-# elif defined (__APPLE__)
-
-  {
-    off_t sbytes = count;
-    res = sendfile (ifd, ofd, offset, &sbytes, 0, 0);
-
-    /* according to the manpage, sbytes is always valid */
-    if (sbytes)
+    if (res < 0 && sbytes)
+      /* maybe only on EAGAIN: as usual, the manpage leaves you guessing */
       res = sbytes;
   }
 
@@ -997,16 +878,6 @@ eio__sendfile (int ofd, int ifd, off_t offset, size_t count, etp_worker *self)
   }
 
 # endif
-
-#elif defined (_WIN32)
-
-  /* does not work, just for documentation of what would need to be done */
-  {
-    HANDLE h = TO_SOCKET (ifd);
-    SetFilePointer (h, offset, 0, FILE_BEGIN);
-    res = TransmitFile (TO_SOCKET (ofd), h, count, 0, 0, 0, 0);
-  }
-
 #else
   res = -1;
   errno = ENOSYS;
@@ -1014,11 +885,6 @@ eio__sendfile (int ofd, int ifd, off_t offset, size_t count, etp_worker *self)
 
   if (res <  0
       && (errno == ENOSYS || errno == EINVAL || errno == ENOTSOCK
-          /* BSDs */
-#ifdef ENOTSUP /* sigh, if the steenking pile called openbsd would only try to at least compile posix code... */
-          || errno == ENOTSUP
-#endif
-          || errno == EOPNOTSUPP /* BSDs */
 #if __solaris
           || errno == EAFNOSUPPORT || errno == EPROTOTYPE
 #endif
@@ -1059,171 +925,6 @@ eio__sendfile (int ofd, int ifd, off_t offset, size_t count, etp_worker *self)
   return res;
 }
 
-static signed char
-eio_dent_cmp (const eio_dirent *a, const eio_dirent *b)
-{
-  return a->score - b->score ? a->score - b->score /* works because our signed char is always 0..100 */
-       : a->inode < b->inode ? -1
-       : a->inode > b->inode ?  1
-       :                        0;
-}
-
-#define EIO_DENT_CMP(i,op,j) eio_dent_cmp (&i, &j) op 0
-
-#define EIO_SORT_CUTOFF 30 /* quite high, but performs well on many filesystems */
-#define EIO_SORT_FAST   60 /* when to only use insertion sort */
-
-static void
-eio_dent_radix_sort (eio_dirent *dents, int size, signed char score_bits, ino_t inode_bits)
-{
-  unsigned char bits [9 + sizeof (ino_t) * 8];
-  unsigned char *bit = bits;
-
-  assert (CHAR_BIT == 8);
-  assert (sizeof (eio_dirent) * 8 < 256);
-  assert (offsetof (eio_dirent, inode)); /* we use bit #0 as sentinel */
-  assert (offsetof (eio_dirent, score)); /* we use bit #0 as sentinel */
-
-  if (size <= EIO_SORT_FAST)
-    return;
-
-  /* first prepare an array of bits to test in our radix sort */
-  /* try to take endianness into account, as well as differences in ino_t sizes */
-  /* inode_bits must contain all inodes ORed together */
-  /* which is used to skip bits that are 0 everywhere, which is very common */
-  {
-    ino_t endianness;
-    int i, j;
-
-    /* we store the byte offset of byte n into byte n of "endianness" */
-    for (i = 0; i < sizeof (ino_t); ++i)
-      ((unsigned char *)&endianness)[i] = i;
-
-    *bit++ = 0;
-
-    for (i = 0; i < sizeof (ino_t); ++i)
-      {
-        /* shifting off the byte offsets out of "endianness" */
-        int offs = (offsetof (eio_dirent, inode) + (endianness & 0xff)) * 8;
-        endianness >>= 8;
-
-        for (j = 0; j < 8; ++j)
-          if (inode_bits & (((ino_t)1) << (i * 8 + j)))
-            *bit++ = offs + j;
-      }
-
-    for (j = 0; j < 8; ++j)
-      if (score_bits & (1 << j))
-        *bit++ = offsetof (eio_dirent, score) * 8 + j;
-  }
-
-  /* now actually do the sorting (a variant of MSD radix sort) */
-  {
-    eio_dirent    *base_stk [9 + sizeof (ino_t) * 8], *base;
-    eio_dirent    *end_stk  [9 + sizeof (ino_t) * 8], *end;
-    unsigned char *bit_stk  [9 + sizeof (ino_t) * 8];
-    int stk_idx = 0;
-
-    base_stk [stk_idx] = dents;
-    end_stk  [stk_idx] = dents + size;
-    bit_stk  [stk_idx] = bit - 1;
-
-    do
-      {
-        base = base_stk [stk_idx];
-        end  = end_stk  [stk_idx];
-        bit  = bit_stk  [stk_idx];
-
-        for (;;)
-          {
-            unsigned char O = *bit >> 3;
-            unsigned char M = 1 << (*bit & 7);
-
-            eio_dirent *a = base;
-            eio_dirent *b = end;
-
-            if (b - a < EIO_SORT_CUTOFF)
-              break;
-
-            /* now bit-partition the array on the bit */
-            /* this ugly asymmetric loop seems to perform much better than typical */
-            /* partition algos found in the literature */
-            do
-              if (!(((unsigned char *)a)[O] & M))
-                ++a;
-              else if (!(((unsigned char *)--b)[O] & M))
-                {
-                  eio_dirent tmp = *a; *a = *b; *b = tmp;
-                  ++a;
-                }
-            while (b > a);
-
-            /* next bit, or stop, if no bits left in this path */
-            if (!*--bit)
-              break;
-
-            base_stk [stk_idx] = a;
-            end_stk  [stk_idx] = end;
-            bit_stk  [stk_idx] = bit;
-            ++stk_idx;
-
-            end = a;
-          }
-      }
-    while (stk_idx--);
-  }
-}
-
-static void
-eio_dent_insertion_sort (eio_dirent *dents, int size)
-{
-  /* first move the smallest element to the front, to act as a sentinel */
-  {
-    int i;
-    eio_dirent *min = dents;
-    
-    /* the radix pre-pass ensures that the minimum element is in the first EIO_SORT_CUTOFF + 1 elements */
-    for (i = size > EIO_SORT_FAST ? EIO_SORT_CUTOFF + 1 : size; --i; )
-      if (EIO_DENT_CMP (dents [i], <, *min))
-        min = &dents [i];
-
-    /* swap elements 0 and j (minimum) */
-    {
-      eio_dirent tmp = *dents; *dents = *min; *min = tmp;
-    }
-  }
-
-  /* then do standard insertion sort, assuming that all elements are >= dents [0] */
-  {
-    eio_dirent *i, *j;
-
-    for (i = dents + 1; i < dents + size; ++i)
-      {
-        eio_dirent value = *i;
-
-        for (j = i - 1; EIO_DENT_CMP (*j, >, value); --j)
-          j [1] = j [0];
-
-        j [1] = value;
-      }
-  }
-}
-
-static void
-eio_dent_sort (eio_dirent *dents, int size, signed char score_bits, ino_t inode_bits)
-{
-  if (size <= 1)
-    return; /* our insertion sort relies on size > 0 */
-
-  /* first we use a radix sort, but only for dirs >= EIO_SORT_FAST */
-  /* and stop sorting when the partitions are <= EIO_SORT_CUTOFF */
-  eio_dent_radix_sort (dents, size, score_bits, inode_bits);
-
-  /* use an insertion sort at the end, or for small arrays, */
-  /* as insertion sort is more efficient for small partitions */
-  eio_dent_insertion_sort (dents, size);
-}
-
 /* read a full directory */
 static void
 eio__scandir (eio_req *req, etp_worker *self)
@@ -1231,306 +932,81 @@ eio__scandir (eio_req *req, etp_worker *self)
   DIR *dirp;
   EIO_STRUCT_DIRENT *entp;
   char *name, *names;
-  int namesalloc = 4096;
-  int namesoffs = 0;
-  int flags = req->int1;
-  eio_dirent *dents = 0;
-  int dentalloc = 128;
-  int dentoffs = 0;
-  ino_t inode_bits = 0;
-
-  req->result = -1;
-
-  if (!(flags & EIO_READDIR_DENTS))
-    flags &= ~(EIO_READDIR_DIRS_FIRST | EIO_READDIR_STAT_ORDER);
+  int memlen = 4096;
+  int memofs = 0;
+  int res = 0;
 
   X_LOCK (wrklock);
   /* the corresponding closedir is in ETP_WORKER_CLEAR */
   self->dirp = dirp = opendir (req->ptr1);
-
-  req->flags |= EIO_FLAG_PTR1_FREE | EIO_FLAG_PTR2_FREE;
-  req->ptr1 = dents = flags ? malloc (dentalloc * sizeof (eio_dirent)) : 0;
-  req->ptr2 = names = malloc (namesalloc);
+  req->flags |= EIO_FLAG_PTR2_FREE;
+  req->ptr2 = names = malloc (memlen);
   X_UNLOCK (wrklock);
 
-  if (dirp && names && (!flags || dents))
+  if (dirp && names)
     for (;;)
       {
         errno = 0;
         entp = readdir (dirp);
 
         if (!entp)
-          {
-            if (errno)
-              break;
+          break;
 
-            /* sort etc. */
-            req->int1   = flags;
-            req->result = dentoffs;
-
-            if (flags & EIO_READDIR_STAT_ORDER)
-              eio_dent_sort (dents, dentoffs, flags & EIO_READDIR_DIRS_FIRST ? 7 : 0, inode_bits);
-            else if (flags & EIO_READDIR_DIRS_FIRST)
-              if (flags & EIO_READDIR_FOUND_UNKNOWN)
-                eio_dent_sort (dents, dentoffs, 7, inode_bits); /* sort by score and inode */
-              else
-                {
-                  /* in this case, all is known, and we just put dirs first and sort them */
-                  eio_dirent *oth = dents + dentoffs;
-                  eio_dirent *dir = dents;
-
-                  /* now partition dirs to the front, and non-dirs to the back */
-                  /* by walking from both sides and swapping if necessary */
-                  while (oth > dir)
-                    {
-                      if (dir->type == EIO_DT_DIR)
-                        ++dir;
-                      else if ((--oth)->type == EIO_DT_DIR)
-                        {
-                          eio_dirent tmp = *dir; *dir = *oth; *oth = tmp;
-
-                          ++dir;
-                        }
-                    }
-
-                  /* now sort the dirs only (dirs all have the same score) */
-                  eio_dent_sort (dents, dir - dents, 0, inode_bits);
-                }
-
-            break;
-          }
-
-        /* now add the entry to our list(s) */
         name = entp->d_name;
 
-        /* skip . and .. entries */
         if (name [0] != '.' || (name [1] && (name [1] != '.' || name [2])))
           {
-            int len = D_NAMLEN (entp) + 1;
+            int len = strlen (name) + 1;
 
-            while (expect_false (namesoffs + len > namesalloc))
+            res++;
+
+            while (memofs + len > memlen)
               {
-                namesalloc *= 2;
+                memlen *= 2;
                 X_LOCK (wrklock);
-                req->ptr2 = names = realloc (names, namesalloc);
+                req->ptr2 = names = realloc (names, memlen);
                 X_UNLOCK (wrklock);
 
                 if (!names)
                   break;
               }
 
-            memcpy (names + namesoffs, name, len);
-
-            if (dents)
-              {
-                struct eio_dirent *ent;
-
-                if (expect_false (dentoffs == dentalloc))
-                  {
-                    dentalloc *= 2;
-                    X_LOCK (wrklock);
-                    req->ptr1 = dents = realloc (dents, dentalloc * sizeof (eio_dirent));
-                    X_UNLOCK (wrklock);
-
-                    if (!dents)
-                      break;
-                  }
-
-                ent = dents + dentoffs;
-
-                ent->nameofs = namesoffs; /* rather dirtily we store the offset in the pointer */
-                ent->namelen = len - 1;
-                ent->inode   = D_INO (entp);
-
-                inode_bits |= ent->inode;
-
-                switch (D_TYPE (entp))
-                  {
-                    default:
-                      ent->type = EIO_DT_UNKNOWN;
-                      flags |= EIO_READDIR_FOUND_UNKNOWN;
-                      break;
-
-                    #ifdef DT_FIFO
-                      case DT_FIFO: ent->type = EIO_DT_FIFO; break;
-                    #endif
-                    #ifdef DT_CHR
-                      case DT_CHR:  ent->type = EIO_DT_CHR;  break;
-                    #endif          
-                    #ifdef DT_MPC
-                      case DT_MPC:  ent->type = EIO_DT_MPC;  break;
-                    #endif          
-                    #ifdef DT_DIR
-                      case DT_DIR:  ent->type = EIO_DT_DIR;  break;
-                    #endif          
-                    #ifdef DT_NAM
-                      case DT_NAM:  ent->type = EIO_DT_NAM;  break;
-                    #endif          
-                    #ifdef DT_BLK
-                      case DT_BLK:  ent->type = EIO_DT_BLK;  break;
-                    #endif          
-                    #ifdef DT_MPB
-                      case DT_MPB:  ent->type = EIO_DT_MPB;  break;
-                    #endif          
-                    #ifdef DT_REG
-                      case DT_REG:  ent->type = EIO_DT_REG;  break;
-                    #endif          
-                    #ifdef DT_NWK
-                      case DT_NWK:  ent->type = EIO_DT_NWK;  break;
-                    #endif          
-                    #ifdef DT_CMP
-                      case DT_CMP:  ent->type = EIO_DT_CMP;  break;
-                    #endif          
-                    #ifdef DT_LNK
-                      case DT_LNK:  ent->type = EIO_DT_LNK;  break;
-                    #endif
-                    #ifdef DT_SOCK
-                      case DT_SOCK: ent->type = EIO_DT_SOCK; break;
-                    #endif
-                    #ifdef DT_DOOR
-                      case DT_DOOR: ent->type = EIO_DT_DOOR; break;
-                    #endif
-                    #ifdef DT_WHT
-                      case DT_WHT:  ent->type = EIO_DT_WHT;  break;
-                    #endif
-                  }
-
-                ent->score = 7;
-
-                if (flags & EIO_READDIR_DIRS_FIRST)
-                  {
-                    if (ent->type == EIO_DT_UNKNOWN)
-                      {
-                        if (*name == '.') /* leading dots are likely directories, and, in any case, rare */
-                          ent->score = 1;
-                        else if (!strchr (name, '.')) /* absense of dots indicate likely dirs */
-                          ent->score = len <= 2 ? 4 - len : len <= 4 ? 4 : len <= 7 ? 5 : 6; /* shorter == more likely dir, but avoid too many classes */
-                      }
-                    else if (ent->type == EIO_DT_DIR)
-                      ent->score = 0;
-                  }
-              }
-
-            namesoffs += len;
-            ++dentoffs;
-          }
-
-        if (EIO_CANCELLED (req))
-          {
-            errno = ECANCELED;
-            break;
+            memcpy (names + memofs, name, len);
+            memofs += len;
           }
       }
+
+  if (errno)
+    res = -1;
+  
+  req->result = res;
 }
 
-#ifdef PAGESIZE
-# define eio_pagesize() PAGESIZE
-#else
-static intptr_t
-eio_pagesize (void)
-{
-  static intptr_t page;
-
-  if (!page)
-    page = sysconf (_SC_PAGESIZE);
-
-  return page;
-}
-#endif
-
-static void
-eio_page_align (void **addr, size_t *length)
-{
-  intptr_t mask = eio_pagesize () - 1;
-
-  /* round down addr */
-  intptr_t adj = mask & (intptr_t)*addr;
-
-  *addr   = (void *)((intptr_t)*addr - adj);
-  *length += adj;
-
-  /* round up length */
-  *length = (*length + mask) & ~mask;
-}
-
-#if !_POSIX_MEMLOCK
-# define eio__mlockall(a) ((errno = ENOSYS), -1)
-#else
-
-static int
-eio__mlockall (int flags)
-{
-  #if __GLIBC__ == 2 && __GLIBC_MINOR__ <= 7
-    extern int mallopt (int, int);
-    mallopt (-6, 238); /* http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=473812 */
-  #endif
-
-  if (EIO_MCL_CURRENT   != MCL_CURRENT
-      || EIO_MCL_FUTURE != MCL_FUTURE)
-    {
-      flags = 0
-         | (flags & EIO_MCL_CURRENT ? MCL_CURRENT : 0)
-         | (flags & EIO_MCL_FUTURE  ? MCL_FUTURE : 0);
-    }
-
-  return mlockall (flags);
-}
-#endif
-
-#if !_POSIX_MEMLOCK_RANGE
-# define eio__mlock(a,b) ((errno = ENOSYS), -1)
-#else
-
-static int
-eio__mlock (void *addr, size_t length)
-{
-  eio_page_align (&addr, &length);
-
-  return mlock (addr, length);
-}
-
-#endif
-
 #if !(_POSIX_MAPPED_FILES && _POSIX_SYNCHRONIZED_IO)
-# define eio__msync(a,b,c) ((errno = ENOSYS), -1)
-#else
-
-int
-eio__msync (void *mem, size_t len, int flags)
-{
-  eio_page_align (&mem, &len);
-
-  if (EIO_MS_ASYNC         != MS_SYNC
-      || EIO_MS_INVALIDATE != MS_INVALIDATE
-      || EIO_MS_SYNC       != MS_SYNC)
-    {
-      flags = 0
-         | (flags & EIO_MS_ASYNC      ? MS_ASYNC : 0)
-         | (flags & EIO_MS_INVALIDATE ? MS_INVALIDATE : 0)
-         | (flags & EIO_MS_SYNC       ? MS_SYNC : 0);
-    }
-
-  return msync (mem, len, flags);
-}
-
+# define msync(a,b,c) ENOSYS
 #endif
 
 int
 eio__mtouch (void *mem, size_t len, int flags)
 {
-  eio_page_align (&mem, &len);
+  intptr_t addr = (intptr_t)mem;
+  intptr_t end = addr + len;
+#ifdef PAGESIZE
+  const intptr_t page = PAGESIZE;
+#else
+  static intptr_t page;
 
-  {
-    intptr_t addr = (intptr_t)mem;
-    intptr_t end = addr + len;
-    intptr_t page = eio_pagesize ();
+  if (!page)
+    page = sysconf (_SC_PAGESIZE);
+#endif
 
-    if (addr < end)
-      if (flags & EIO_MT_MODIFY) /* modify */
-        do { *((volatile sig_atomic_t *)addr) |= 0; } while ((addr += page) < len);
-      else
-        do { *((volatile sig_atomic_t *)addr)     ; } while ((addr += page) < len);
-  }
+  addr &= ~(page - 1); /* assume page size is always a power of two */
+
+  if (addr < end)
+    if (flags) /* modify */
+      do { *((volatile sig_atomic_t *)addr) |= 0; } while ((addr += page) < len);
+    else
+      do { *((volatile sig_atomic_t *)addr)     ; } while ((addr += page) < len);
 
   return 0;
 }
@@ -1574,7 +1050,7 @@ X_THREAD_PROC (etp_proc)
 
           ++idle;
 
-          ts.tv_sec = time (0) + idle_timeout;
+          ts.tv_sec = time (0) + IDLE_TIMEOUT;
           if (X_COND_TIMEDWAIT (reqwait, reqlock, ts) == ETIMEDOUT)
             {
               if (idle > max_idle)
@@ -1663,6 +1139,8 @@ static void eio_api_destroy (eio_req *req)
 
 static void eio_execute (etp_worker *self, eio_req *req)
 {
+  errno = 0;
+
   switch (req->type)
     {
       case EIO_READ:      ALLOC (req->size);
@@ -1683,11 +1161,6 @@ static void eio_execute (etp_worker *self, eio_req *req)
       case EIO_FSTAT:     ALLOC (sizeof (EIO_STRUCT_STAT));
                           req->result = fstat     (req->int1, (EIO_STRUCT_STAT *)req->ptr2); break;
 
-      case EIO_STATVFS:   ALLOC (sizeof (EIO_STRUCT_STATVFS));
-                          req->result = statvfs   (req->ptr1, (EIO_STRUCT_STATVFS *)req->ptr2); break;
-      case EIO_FSTATVFS:  ALLOC (sizeof (EIO_STRUCT_STATVFS));
-                          req->result = fstatvfs  (req->int1, (EIO_STRUCT_STATVFS *)req->ptr2); break;
-
       case EIO_CHOWN:     req->result = chown     (req->ptr1, req->int2, req->int3); break;
       case EIO_FCHOWN:    req->result = fchown    (req->int1, req->int2, req->int3); break;
       case EIO_CHMOD:     req->result = chmod     (req->ptr1, (mode_t)req->int2); break;
@@ -1704,31 +1177,28 @@ static void eio_execute (etp_worker *self, eio_req *req)
       case EIO_RENAME:    req->result = rename    (req->ptr1, req->ptr2); break;
       case EIO_LINK:      req->result = link      (req->ptr1, req->ptr2); break;
       case EIO_SYMLINK:   req->result = symlink   (req->ptr1, req->ptr2); break;
-      case EIO_MKNOD:     req->result = mknod     (req->ptr1, (mode_t)req->int2, (dev_t)req->offs); break;
+      case EIO_MKNOD:     req->result = mknod     (req->ptr1, (mode_t)req->int2, (dev_t)req->int3); break;
 
-      case EIO_READLINK:  ALLOC (PATH_MAX);
-                          req->result = readlink  (req->ptr1, req->ptr2, PATH_MAX); break;
+      case EIO_READLINK:  ALLOC (NAME_MAX);
+                          req->result = readlink  (req->ptr1, req->ptr2, NAME_MAX); break;
 
       case EIO_SYNC:      req->result = 0; sync (); break;
       case EIO_FSYNC:     req->result = fsync     (req->int1); break;
       case EIO_FDATASYNC: req->result = fdatasync (req->int1); break;
-      case EIO_MSYNC:     req->result = eio__msync (req->ptr2, req->size, req->int1); break;
+      case EIO_MSYNC:     req->result = msync (req->ptr2, req->size, req->int1); break;
       case EIO_MTOUCH:    req->result = eio__mtouch (req->ptr2, req->size, req->int1); break;
-      case EIO_MLOCK:     req->result = eio__mlock (req->ptr2, req->size); break;
-      case EIO_MLOCKALL:  req->result = eio__mlockall (req->int1); break;
-      case EIO_SYNC_FILE_RANGE: req->result = eio__sync_file_range (req->int1, req->offs, req->size, req->int2); break;
 
       case EIO_READDIR:   eio__scandir (req, self); break;
 
       case EIO_BUSY:
 #ifdef _WIN32
-	Sleep (req->nv1 * 1e3);
+	Sleep (req->nv1 * 1000.);
 #else
         {
           struct timeval tv;
 
           tv.tv_sec  = req->nv1;
-          tv.tv_usec = (req->nv1 - tv.tv_sec) * 1e6;
+          tv.tv_usec = (req->nv1 - tv.tv_sec) * 1000000.;
 
           req->result = select (0, 0, 0, 0, &tv);
         }
@@ -1753,6 +1223,7 @@ static void eio_execute (etp_worker *self, eio_req *req)
           else
             times = 0;
 
+
           req->result = req->type == EIO_FUTIME
                         ? futimes (req->int1, times)
                         : utimes  (req->ptr1, times);
@@ -1771,7 +1242,6 @@ static void eio_execute (etp_worker *self, eio_req *req)
         break;
 
       default:
-        errno = ENOSYS;
         req->result = -1;
         break;
     }
@@ -1811,21 +1281,6 @@ eio_req *eio_mtouch (void *addr, size_t length, int flags, int pri, eio_cb cb, v
   REQ (EIO_MTOUCH); req->ptr2 = addr; req->size = length; req->int1 = flags; SEND;
 }
 
-eio_req *eio_mlock (void *addr, size_t length, int pri, eio_cb cb, void *data)
-{
-  REQ (EIO_MLOCK); req->ptr2 = addr; req->size = length; SEND;
-}
-
-eio_req *eio_mlockall (int flags, int pri, eio_cb cb, void *data)
-{
-  REQ (EIO_MLOCKALL); req->int1 = flags; SEND;
-}
-
-eio_req *eio_sync_file_range (int fd, off_t offset, size_t nbytes, unsigned int flags, int pri, eio_cb cb, void *data)
-{
-  REQ (EIO_SYNC_FILE_RANGE); req->int1 = fd; req->offs = offset; req->size = nbytes; req->int2 = flags; SEND;
-}
-
 eio_req *eio_fdatasync (int fd, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FDATASYNC); req->int1 = fd; SEND;
@@ -1856,11 +1311,6 @@ eio_req *eio_fstat (int fd, int pri, eio_cb cb, void *data)
   REQ (EIO_FSTAT); req->int1 = fd; SEND;
 }
 
-eio_req *eio_fstatvfs (int fd, int pri, eio_cb cb, void *data)
-{
-  REQ (EIO_FSTATVFS); req->int1 = fd; SEND;
-}
-
 eio_req *eio_futime (int fd, double atime, double mtime, int pri, eio_cb cb, void *data)
 {
   REQ (EIO_FUTIME); req->int1 = fd; req->nv1 = atime; req->nv2 = mtime; SEND;
@@ -1942,11 +1392,6 @@ eio_req *eio_lstat (const char *path, int pri, eio_cb cb, void *data)
   return eio__1path (EIO_LSTAT, path, pri, cb, data);
 }
 
-eio_req *eio_statvfs (const char *path, int pri, eio_cb cb, void *data)
-{
-  return eio__1path (EIO_STATVFS, path, pri, cb, data);
-}
-
 eio_req *eio_unlink (const char *path, int pri, eio_cb cb, void *data)
 {
   return eio__1path (EIO_UNLINK, path, pri, cb, data);
@@ -1957,14 +1402,14 @@ eio_req *eio_rmdir (const char *path, int pri, eio_cb cb, void *data)
   return eio__1path (EIO_RMDIR, path, pri, cb, data);
 }
 
-eio_req *eio_readdir (const char *path, int flags, int pri, eio_cb cb, void *data)
+eio_req *eio_readdir (const char *path, int pri, eio_cb cb, void *data)
 {
-  REQ (EIO_READDIR); PATH; req->int1 = flags; SEND;
+  return eio__1path (EIO_READDIR, path, pri, cb, data);
 }
 
 eio_req *eio_mknod (const char *path, mode_t mode, dev_t dev, int pri, eio_cb cb, void *data)
 {
-  REQ (EIO_MKNOD); PATH; req->int2 = (long)mode; req->offs = (off_t)dev; SEND;
+  REQ (EIO_MKNOD); PATH; req->int2 = (long)mode; req->int3 = (long)dev; SEND;
 }
 
 static eio_req *
@@ -2058,15 +1503,12 @@ void eio_grp_add (eio_req *grp, eio_req *req)
 ssize_t eio_sendfile_sync (int ofd, int ifd, off_t offset, size_t count)
 {
   etp_worker wrk;
-  ssize_t ret;
 
   wrk.dbuf = 0;
 
-  ret = eio__sendfile (ofd, ifd, offset, count, &wrk);
+  eio__sendfile (ofd, ifd, offset, count, &wrk);
 
   if (wrk.dbuf)
     free (wrk.dbuf);
-
-  return ret;
 }
 
